local ADDON_NAME, addon = ...
_G[ADDON_NAME] = addon

addon.events = addon.events or {}
addon.events.units = {}
addon.db = addon.db or {}
addon.filter = addon.filter or {}
--addon.filter.hide = {}
addon.filter.color = {}
addon.scheduler = addon.scheduler or {}
addon.scheduler.tasks = {}

local db = addon.db

addon.defaults = {
	version = 0,
	nameFont = {name = nil, size = 11.5, outline = "OUTLINE"},
	numFont = {name = nil, size = 8.5, outline = "OUTLINE"},
	castFont = {name = nil, size = 9.5, outline = "OUTLINE"},
	frameSize = {W = 112, H = 9},
	frameScale = {x = 1.05, y = 1},
	healthBarHeight = 7,
	castBarHeight = 6,
	frameMargin = {y = 30},
	barTexture = "Interface\\Addons\\NamePlateKAI\\texture\\normTex",
	castBarTexture = nil,
	targetAlwaysOnTop = true,
	redHostile = false,
	disableFactionIcon = false,
	disableDefaultFilters = false,
	friendlyClassColor = false,
	barBackdrop = {
		bgFile = nil,
		edgeFile = "Interface\\AddOns\\NamePlateKAI\\texture\\glowTex",
		--edgeFile = "Interface\\AddOns\\NamePlateKAI\\texture\\border_sharp",
		tile = false,
		edgeSize = 4,
		insets = {left = 0, right = 0, top = 0, bottom = 0}
	},
	castbarBackdrop = {
		bgFile = nil,
		edgeFile = "Interface\\AddOns\\NamePlateKAI\\texture\\glowTex",
		tile = false,
		edgeSize = 4,
		insets = nil,
	},
	--borderColor = {0, 0, 0, 0.7},
	barColors = {
		friendly = {0.225, 0.415, 0.68},
		npc = {0.45, 0.62, 0.45},
		hostile = {0.27, 0.72, 0.27},
		neutral = {0.55, 0.55, 0.35},
		yellow = {0.8, 0.7, 0},
		orange = {0.9, 0.3, 0},
		red = {0.8, 0, 0},
		tapped = {0.37, 0.45, 0.50},
	},
	spellIcon = {size = 16},
	raidIcon = {size = 16},
	filterColors = {
		[1] = {1, 1, 0.65},
		[2] = {1, 0.72, 1},
		[3] = {0.3, 0.6, 1},
		[4] = {1, 0.1, 1, 1}, -- no threat (purple)
		[5] = {0.45, 0.5, 0.55, 1}, -- trash mob
		[6] = {0.45, 0.85, 0.85, 1}, -- no threat (light blue)
		[7] = {0.6, 0.55, 0.5} -- normal mob
	},
	target = {
		flash = true,
		shine = true,
		shineAlpha = 0.8,
		animation = true,
		animationAlpha = 0.3,
		animationStyle = "default",
		animationOverlay = false,
	},
	healthText = false,
}

addon.defaults.barColors[1] = addon.defaults.barColors.hostile
addon.defaults.barColors[2] = addon.defaults.barColors.hostile
addon.defaults.barColors[3] = addon.defaults.barColors.orange
addon.defaults.barColors[4] = addon.defaults.barColors.neutral
addon.defaults.barColors[5] = addon.defaults.barColors.npc
addon.defaults.barColors[6] = addon.defaults.barColors.npc
addon.defaults.barColors[7] = addon.defaults.barColors.npc
addon.defaults.barColors[8] = addon.defaults.barColors.npc
addon.defaults.barColors[9] = addon.defaults.barColors.friendly
addon.defaults.barColors[10] = addon.defaults.barColors.tapped
addon.defaults.barColors[-1] = addon.defaults.filterColors[5]

addon.defaultFilters = {}

local select, unpack, strfind, tinsert, tremove, tonumber, floor, pairs = select, unpack, string.find, tinsert, tremove, tonumber, math.floor, pairs
local ShouldShowName, UnitLevel, UnitIsUnit, UnitIsPlayer, UnitCanAttack, UnitPlayerControlled, UnitExists, UnitFactionGroup, UnitClass = ShouldShowName, UnitLevel, UnitIsUnit, UnitIsPlayer, UnitCanAttack, UnitPlayerControlled, UnitExists, UnitFactionGroup, UnitClass
local GetNamePlates, GetNamePlateForUnit = C_NamePlate.GetNamePlates, C_NamePlate.GetNamePlateForUnit

local FontUnitNameNormal = CreateFont(ADDON_NAME.."NameNormal")
local FontUnitNameSmall = CreateFont(ADDON_NAME.."NameGrouped")
local FontNumber = CreateFont(ADDON_NAME.."Number")
local FontSpellName = CreateFont(ADDON_NAME.."SpellName")

addon.frame = addon.frame or CreateFrame("Frame", nil, _G.WorldFrame)

local ParentFrame1 = CreateFrame("Frame", nil, _G.WorldFrame)
ParentFrame1:SetFrameStrata("BACKGROUND")
ParentFrame1:SetFrameLevel(2)

local ParentFrame2 = CreateFrame("Frame", nil, _G.WorldFrame)
ParentFrame2:SetFrameStrata("LOW")

local ParentFrame3 = CreateFrame("Frame", nil, _G.WorldFrame)
ParentFrame3:SetFrameStrata("MEDIUM")

local addedPlates = {}
local npcUnitFrames = {}

local GetCreatureTitle do
	local tooltip = CreateFrame("GameTooltip")
	tooltip.leftLines = {}
	for i = 1, 2 do
		local L,R = tooltip:CreateFontString(), tooltip:CreateFontString()
		L:SetFontObject(GameFontNormal)
		R:SetFontObject(GameFontNormal)
		tooltip:AddFontStrings(L,R)
		tooltip.leftLines[i] = L
	end
	tooltip:SetOwner(_G.WorldFrame, "ANCHOR_NONE")

	local FORMAT_TOOLTIP_LEVEL = "^"..TOOLTIP_UNIT_LEVEL:gsub("%%s", ".+")
	GetCreatureTitle = function(unitID)
		tooltip:ClearLines()
		tooltip:SetUnit(unitID)
		local text = tooltip.leftLines[2]:GetText()
		--tooltip:Hide()
		if text and not strfind(text, FORMAT_TOOLTIP_LEVEL) then
			return text
		end
	end
end

local UnitFactionIcon do
	local icons = {Horde = 130705, Alliance = 130704, Neutral = 1140616}
	UnitFactionIcon = function(unitID)
		return icons[UnitFactionGroup(unitID) or 1]
	end
end

local function UpdateStatusText(UnitFrame, titleText, noText)
	if UnitFrame.isFriendlyNPC and UnitFrame.unitExists then
		local isTarget = addon.targetPlate and addon.targetPlate.UnitFrame == UnitFrame
		local valid = ShouldShowName(UnitFrame)
		if not noText then
			titleText = titleText or isTarget or level == -1
			local text
			if valid or titleText then
				local level = UnitLevel(UnitFrame.unit)
				text = titleText and GetCreatureTitle(UnitFrame.unit) or (level == -1) and _G.BOSS or level and _G.UNIT_LEVEL_TEMPLATE:format(level)
			end
			UnitFrame.statusText:SetText(text or nil)
			if titleText then
				npcUnitFrames[UnitFrame] = nil
			end
		end
		UnitFrame.statusText:SetShown(valid)
		if isTarget then
			UnitFrame.healthBar:SetAlpha(1)
			UnitFrame.statusText:SetWidth(0)
		else
			UnitFrame.healthBar:SetAlpha(0.4)
			UnitFrame.statusText:SetWidth(addon.statusTextWidth)
		end
		return valid
	else
		UnitFrame.statusText:Hide()
		UnitFrame.healthBar:SetAlpha(1)
	end
end

function NameplateBuffContainerMixin:UpdateAnchor() -- hook default nameplates api
	local targetYOffset = addon.targetPlate and (addon.targetPlate.UnitFrame.BuffFrame == self) and (self:GetTargetYOffset() + 4) or 0.0
	if self:GetParent().unit and ShouldShowName(self:GetParent()) then
		targetYOffset = targetYOffset + db.nameFont.size
	end
	self:SetPoint("BOTTOM", self:GetParent().healthBar, "TOP", 0, 4 + targetYOffset);
end

local function NamePlate_OnSizeChanged(namePlate)
	if namePlate.fadeIn then
		if namePlate:GetScale() > 0.79 then
			namePlate.fadeIn = false
		end
	elseif namePlate.fadeOut then
		namePlate.UnitFrame:SetAlpha(namePlate:GetScale())
	elseif namePlate:GetScale() < 0.79 then
		namePlate.fadeOut = true
		namePlate.UnitFrame:SetAlpha(namePlate:GetScale())
	end
end

local function SetKaiBackdrop(frame, backdrop)
	frame:SetBackdrop(backdrop)
	local offset = backdrop.edgeSize
	local regs = {frame:GetRegions()}
	local v, p, r, rp, x, y
	for i = 1, #regs do
		v = regs[i]
		if v:GetObjectType() == "Texture" and v:GetTexture() == backdrop.edgeFile then
			for j = 1, v:GetNumPoints() do
				p, r, rp, x, y = v:GetPoint(j)
				if x and y then
					if strfind(p, "TOP", 1, true) then y = y + offset end
					if strfind(p, "BOTTOM", 1, true) then y = y - offset end
					if strfind(p, "LEFT", 1, true) then x = x - offset end
					if strfind(p, "RIGHT", 1, true) then x = x + offset end
					v:SetPoint(p, r, rp, x, y)
				end
			end
		end
	end
	wipe(regs)
end

local function CastBar_OnShow(castBar)
	castBar.Icon:Show()
end

function addon:NamePlateOnCreate(namePlate)
	local UnitFrame = namePlate.UnitFrame
	local healthBar = UnitFrame.healthBar

	do
		local frame = namePlate.UnitFrame.ClassificationFrame
		namePlate:SetScript("OnUpdate", function()
			frame:ClearAllPoints()
			frame:SetPoint("RIGHT", healthBar, "LEFT", 1, 0)
		end)
	end

	--UnitFrame.ClassificationFrame:SetSize(16, 14)
	UnitFrame.ClassificationFrame:SetFrameLevel(2000)
	--UnitFrame.RaidTargetFrame:SetFrameStrata("LOW")
	UnitFrame.RaidTargetFrame:SetFrameLevel(2000)
	--UnitFrame.RaidTargetFrame.RaidTargetIcon:SetDrawLayer("ARTWORK", 6)
	UnitFrame.RaidTargetFrame:ClearAllPoints()
	UnitFrame.RaidTargetFrame:SetPoint("CENTER", healthBar, "CENTER", 0, 0)
	UnitFrame.RaidTargetFrame:SetSize(db.raidIcon.size, db.raidIcon.size)

	UnitFrame.statusText:ClearAllPoints()
	UnitFrame.statusText:SetPoint("TOP", healthBar, "TOP", 0, 2)
	UnitFrame.statusText:SetFontObject(FontUnitNameSmall)
	UnitFrame.statusText:SetTextColor(0.9, 0.9, 0.7, 1)
	UnitFrame.statusText:SetParent(UnitFrame.RaidTargetFrame)
	UnitFrame.statusText:SetWidth(self.statusTextWidth)
	UnitFrame.statusText:SetWordWrap(false)

	UnitFrame.name:SetParent(UnitFrame.RaidTargetFrame)

	UnitFrame.selectionHighlight:SetTexture(nil)

	healthBar.fill = healthBar:CreateTexture()
	healthBar.fill:SetTexture(db.barTexture)
	healthBar.fill:SetAllPoints(healthBar.barTexture)

	healthBar.border:Hide()
	for i = 1, #healthBar.border.Textures do
		healthBar.border.Textures[i]:SetTexture(nil)
	end
	wipe(healthBar.border.Textures)

	SetKaiBackdrop(healthBar, db.barBackdrop)
	healthBar.background:SetTexture(nil)
	--healthBar.background:SetVertexColor(0.1, 0.1, 0.1, 1)
	healthBar:SetBackdropColor(0.1, 0.1, 0.1, 0.85)

	UnitFrame.castBar:SetScript("OnShow", CastBar_OnShow)
	--UnitFrame.castBar.Icon = UnitFrame.castBar:CreateTexture()

	namePlate:SetScript("OnSizeChanged", NamePlate_OnSizeChanged)
end

local GetNpcID do
	local UnitGUID, strmatch = UnitGUID, string.match
	local FORMAT_GUID_CREATURE = "^%a+%-%d+%-%d+%-%d+%-%d+%-(%d+)"
	local id
	function GetNpcID(unitID)
		id = UnitGUID(unitID)
		if id then
			return tonumber(strmatch(id, FORMAT_GUID_CREATURE))
		end
	end
end

local UnitReaction, UnitName = UnitReaction, UnitName

local function SetClassificationIcon(UnitFrame, unitID, defaultIcon)
	if db.disableFactionIcon and addon.inPvP then return end
	UnitFrame.classificationIndicator:SetTexture(UnitFactionIcon(unitID) or defaultIcon)
	UnitFrame.ClassificationFrame:SetSize(20, 20)
	UnitFrame.classificationIndicator:Show()
end

local ClassColors = setmetatable({}, {__index =
	function(self, key)
		local value
		local c = _G.RAID_CLASS_COLORS[key]
		if c then
			value = {c.r, c.g, c.b}
		else
			value = self["*"]
		end
		self[key] = value
		return value
	end
})
ClassColors["*"] = {0.65, 0.65, 0.65}

function addon:NamePlateOnAdded(namePlate, unitID, refresh)
	local UnitFrame = namePlate.UnitFrame
	local healthBar = UnitFrame.healthBar

	UnitFrame.isPlayer = UnitIsUnit("player", unitID)
	UnitFrame.isPvP = false
	UnitFrame.isFriendlyNPC = false
	UnitFrame.reaction = UnitReaction(unitID, "player") or 10

	UnitFrame.ClassificationFrame:SetSize(16, 14)

	local barColor = self.filter.color[GetNpcID(unitID) or 0] or self.filter.color[UnitName(unitID) or 0]
	if barColor then
		UnitFrame.ignoreThreat = barColor[4] ~= nil
		healthBar.barTexture:SetTexture(nil)
		healthBar.fill:SetVertexColor(unpack(barColor, 1, 3))
		healthBar.fill:Show()
	else
		UnitFrame.ignoreThreat = nil
		if UnitIsPlayer(unitID) then
			UnitFrame.isPvP = true
			if UnitFrame.reaction <= 4 then
				SetClassificationIcon(UnitFrame, unitID, 1140616)
			else
				UnitFrame.reaction = 9
			end
			if UnitFrame.reaction <= 4 and UnitFactionGroup(unitID) or UnitFrame.isPlayer then
				healthBar.barTexture:SetTexture(db.barTexture)
				healthBar.fill:Hide()
			else
				healthBar.barTexture:SetTexture(nil)
				if db.friendlyClassColor then
					local _, class = UnitClass(unitID)
					healthBar.fill:SetVertexColor(unpack(ClassColors[class or ""]))
				else
					healthBar.fill:SetVertexColor(unpack(db.barColors[UnitFrame.reaction], 1, 3))
				end
				healthBar.fill:Show()
			end
		else
			if UnitPlayerControlled(unitID) then
				--UnitFrame.isPvP = true
				if UnitFrame.reaction <= 4 then
					SetClassificationIcon(UnitFrame, unitID)
					UnitFrame.reaction = -1
				end
			else
				UnitFrame.isFriendlyNPC = (UnitFrame.reaction > 2) and not UnitCanAttack("player", unitID)
			end
			healthBar.barTexture:SetTexture(nil)
			healthBar.fill:SetVertexColor(unpack(db.barColors[UnitFrame.reaction], 1, 3))
			healthBar.fill:Show()
		end
	end

	UnitFrame.healthBar:SetBackdropBorderColor(0, 0, 0, 0.8)
	self:UpdateThreatBorder(UnitFrame)

	if UnitFrame.isPlayer then
		UnitFrame:SetParent(namePlate)
		return
	end

	UnitFrame:ClearAllPoints()
	UnitFrame:SetPoint("BOTTOM", namePlate, "BOTTOM", 0, 0)
	UnitFrame:SetSize(self.frameWidth, self.frameHeight)

	local isTarget = self.hasTarget and UnitIsUnit("target", unitID)
	if isTarget then
		self.targetPlate = namePlate
		UnitFrame:SetParent(ParentFrame2)
	else
		UnitFrame:SetParent(ParentFrame1)
	end

	if not UnitFrame.isFriendlyNPC and UnitLevel(unitID) < 0 then
		UnitFrame.classificationIndicator:SetTexture(137025)
		UnitFrame.classificationIndicator:Show()
	end

	local shouldShowTitle = UpdateStatusText(UnitFrame)

	local castBar = UnitFrame.castBar
	castBar.Icon:ClearAllPoints()
	castBar.Icon:SetPoint("BOTTOMLEFT", castBar, "BOTTOMRIGHT", 2, 0)
	castBar.Text:ClearAllPoints()
	castBar.Text:SetPoint("TOP", castBar, "CENTER", 0, 1)

	castBar:GetStatusBarTexture():SetTexture(db.barTexture)
	castBar.Flash:SetTexture(db.castBarTexture or db.barTexture)
	castBar.Icon:SetSize(db.spellIcon.size, db.spellIcon.size)

	if isTarget then
		self:SetTargetGlow()
		npcUnitFrames[UnitFrame] = nil
	elseif shouldShowTitle then
		npcUnitFrames[UnitFrame] = UnitFrame.isFriendlyNPC or nil
	end

	if refresh then return end

	UnitFrame:SetAlpha(0)
	tinsert(addedPlates, namePlate)
	namePlate.fadeIn = true
	namePlate.fadeOut = false
end

function addon:NamePlateOnRemoved(namePlate, unitID)
	--if namePlate.UnitFrame.isPlayer then return end
	namePlate.fadeOut = false

	npcUnitFrames[namePlate.UnitFrame] = nil
	namePlate.UnitFrame:SetParent(namePlate)
	if self.targetPlate == namePlate then
		self.targetPlate = nil
		self:SetTargetGlow()
	end
	if self.frame.highlightFrame == namePlate then
		self.frame.highlight:Hide()
		self.frame.highlightBG:Hide()
		self.frame.highlightFrame = nil
	end
end

function addon:OnTargetChanged(forceNoTarget)
	self.hasTarget = UnitExists("target") and not UnitIsUnit("player", "target")
	self.targetPlate = not forceNoTarget and self.hasTarget and GetNamePlateForUnit("target") or nil
	if self.targetPlate then
		self:RefreshNamePlate(self.targetPlate)
	end

	local UnitFrame
	for k, namePlate in pairs(GetNamePlates()) do
		if namePlate ~= self.targetPlate then
			UnitFrame = namePlate.UnitFrame
			if not UnitFrame.isPlayer then
				UnitFrame:SetParent(ParentFrame1)
				UpdateStatusText(UnitFrame, false, true)
			end
		end
	end

	ParentFrame1:SetAlpha(UnitExists("target") and 0.6 or 1)
	ParentFrame3:SetAlpha(UnitExists("target") and 0.85 or 1)
	if not self.loading then
		self:SetTargetGlow()
	end
end

function addon:NamePlateOnEnter(namePlate)
	if not namePlate.UnitFrame.unitExists then return end
	local UnitFrame = namePlate.UnitFrame
	local unitID = namePlate.namePlateUnitToken
	UpdateStatusText(UnitFrame, true)
	if UnitFrame.isFriendlyNPC and not ShouldShowName(UnitFrame) then
		UnitFrame.statusText:Show()
	end

	if namePlate ~= self.targetPlate then
		UnitFrame:SetParent(ParentFrame3)
	end

	self.frame.highlight:SetPoint("TOPLEFT", UnitFrame.healthBar, "TOPLEFT", -1, 1)
	self.frame.highlight:SetPoint("BOTTOMRIGHT", UnitFrame.healthBar, "BOTTOMRIGHT", 1, -1)
	self.frame.highlight:SetParent(UnitFrame.healthBar)
	self.frame.highlight:Show()
	self.frame.highlight:SetAlpha(0.09)

	self.frame.highlightBG:SetPoint("BOTTOM", UnitFrame.name, "BOTTOM", -1, -1)
	self.frame.highlightBG:SetSize(UnitFrame.name:GetStringWidth(), UnitFrame.name:GetStringHeight())
	self.frame.highlightBG:SetParent(UnitFrame.healthBar)
	self.frame.highlightBG:SetShown(self.targetPlate ~= namePlate and ShouldShowName(UnitFrame))

	self.frame.highlightFrame = namePlate
end

function addon:NamePlateOnLeave(namePlate)
	if namePlate.UnitFrame.isFriendlyNPC and namePlate.UnitFrame.unitExists and not ShouldShowName(namePlate.UnitFrame) then
		namePlate.UnitFrame.statusText:Hide()
	end
	if namePlate ~= self.targetPlate then
		namePlate.UnitFrame:SetParent(ParentFrame1)
	end
end

do
	local UnitThreatSituation = UnitThreatSituation
	local borderColors = {}
	for i = 1, 4 do
		borderColors[i] = {GetThreatStatusColor(i)}
		borderColors[i][4] = 1
	end
	borderColors[0] = {0, 0, 0, 0.8}

	function addon:UpdateThreatBorder(UnitFrame)
		if (UnitFrame.reaction > 4) or UnitFrame.isPvP or UnitFrame.ignoreThreat or UnitFrame.isPlayer then return end
		local status = UnitFrame.unitExists and UnitThreatSituation("player", UnitFrame.unit) or 0
		UnitFrame.healthBar:SetBackdropBorderColor(unpack(borderColors[status]))
	end
end

function addon:QueueSetCVar(key, value)
	if not self.cvarList then
		self.cvarList = {}
	end
	if type(key) == "string" and value then
		self.cvarList[key] = tostring(value)
	end
end

function addon:RestyleNamePlates()
	self:OnTargetChanged(true)
	if InCombatLockdown() then
		if not self.needRestyle then
			self.needRestyle = true
			DEFAULT_CHAT_FRAME:AddMessage("|cff3399ff[NamePlateKAI "..addon.version.."]:|r "..YELLOW_FONT_COLOR_CODE.."Effect appears after combat."..FONT_COLOR_CODE_CLOSE)
		end
	else
		SetCVar("NamePlateHorizontalScale", db.frameScale.x)
		SetCVar("NamePlateVerticalScale", db.frameScale.x)
		NamePlateDriverFrame:UpdateNamePlateOptions(self)
		local width, height = C_NamePlate.GetNamePlateOtherSize()
		C_NamePlate.SetNamePlateOtherSize(width, db.castBarHeight + db.healthBarHeight + db.nameFont.size + db.frameMargin.y)

		if self.cvarList then
			for k, v in pairs(self.cvarList) do
				SetCVar(k, v)
			end
			wipe(self.cvarList)
			self.cvarList = nil
		end
	end

	self.frameWidth, self.frameHeight = C_NamePlate.GetNamePlateOtherSize()
	self.frameWidth, self.frameHeight = floor(self.frameWidth), floor(self.frameHeight)
	self.statusTextWidth = self.frameWidth + 20
	local options = DefaultCompactNamePlateFrameSetUpOptions
	options.healthBarHeight = db.healthBarHeight
	options.castBarHeight = db.castBarHeight
	options.castBarFontHeight = db.castFont.size

	wipe(npcUnitFrames)
	wipe(addedPlates)
	for k, namePlate in ipairs(GetNamePlates()) do
		NamePlateDriverFrame:ApplyFrameOptions(namePlate, namePlate.namePlateUnitToken)
		self:RefreshNamePlate(namePlate)
		namePlate.UnitFrame.BuffFrame:UpdateAnchor()
		namePlate.UnitFrame.RaidTargetFrame:SetSize(db.raidIcon.size, db.raidIcon.size)
	end
	ParentFrame2:SetFrameStrata(db.targetAlwaysOnTop and "MEDIUM" or "LOW")
	ParentFrame3:SetFrameStrata(db.targetAlwaysOnTop and "LOW" or "MEDIUM")
	self:OnTargetChanged()
end

function addon:ThreatUpdateAll()
	for _, namePlate in pairs(GetNamePlates()) do
		self:UpdateThreatBorder(namePlate.UnitFrame)
	end
end

function addon:RefreshNamePlate(namePlate)
	self:NamePlateOnAdded(namePlate, namePlate.namePlateUnitToken, true)
end

function addon:RefreshAll()
	wipe(npcUnitFrames)
	--wipe(addedPlates)
	for _, namePlate in pairs(GetNamePlates()) do
		if namePlate.UnitFrame.unitExists then
			self:NamePlateOnAdded(namePlate, namePlate.namePlateUnitToken, true)
		end
	end
end

do
	local modules = {}

	function addon:EncounterStart(id)
		if modules[id] and not modules[id].enabled then
			if not modules[id].initialized then
				modules[id]:Init(id)
			end
			modules[id].enabled = true
			modules[id]:Enable()
		end
	end

	function addon:EncounterEnd(id)
		if modules[id] and modules[id].enabled then
			modules[id].enabled = nil
			modules[id]:Disable()
		end
	end

	function addon:CheckEncounters()
		if not IsEncounterInProgress() then
			for id, mod in pairs(modules) do
				self:EncounterEnd(id)
			end
		end
	end
end

function addon:UpdateFonts()
	local filename = _G.GameFontWhite:GetFont()
	if not FontUnitNameNormal:SetFont(db.nameFont.name or filename, db.nameFont.size, db.nameFont.outline) then
		FontUnitNameNormal:SetFont(filename, db.nameFont.size, db.nameFont.outline)
	end
	local filename = _G.GameFontWhiteTiny:GetFont()
	if not FontUnitNameSmall:SetFont(db.castFont.name or filename, db.nameFont.size * 0.9, db.castFont.outline) then
		FontUnitNameSmall:SetFont(filename, db.nameFont.size * 0.9, db.castFont.outline)
	end
	if not FontSpellName:SetFont(db.castFont.name or filename, db.castFont.size, db.castFont.outline) then
		FontSpellName:SetFont(filename, db.castFont.size, db.castFont.outline)
	end
	local filename = _G.NumberFont_Shadow_Small:GetFont()
	if not FontNumber:SetFont(db.numFont.name or filename, db.numFont.size, db.numFont.outline) then
		FontNumber:SetFont(filename, db.numFont.size, db.numFont.outline)
	end
end

local function UpdateNamePlateOptions_Handler(self, arg1)
	if arg1 == addon then
		return
	end
	addon:RestyleNamePlates()
end

do
	local empty = {} -- dummy table
	local function tMergeCopy(base, src, dest, exclude)
		src = src or empty
		for k, v in pairs(base) do
			if k == exclude then

			elseif type(v) == "table" then
				if src[k] == nil or type(src[k]) == "table" then
					dest[k] = {}
					tMergeCopy(v, src[k], dest[k])
				else
					dest[k] = src[k]
				end
			else
				dest[k] = src[k]
				if dest[k] == nil then
					dest[k] = v
				end
			end
		end
	end

	local function tShallowCopy(src, dest) for k, v in pairs(src) do dest[k] = v end end

	function addon:LoadProfile(doFilterUpdate)
		local profile = _G.NamePlateKAIDB and _G.NamePlateKAIDB.profiles and _G.NamePlateKAIDB.profiles.Default or nil
		wipe(db)
		tMergeCopy(self.defaults, profile, db, "filters")

		if db.redHostile then
			db.barColors[1] = self.defaults.barColors.red
			db.barColors[2] = self.defaults.barColors.red
		end

		db.barBackdrop.bgFile = db.barBackdrop.bgFile or db.barTexture
		if not db.castbarBackdrop.bgFile then
			db.castbarBackdrop.bgFile = db.barBackdrop.bgFile
		end
		if not db.castbarBackdrop.edgeFile then
			db.castbarBackdrop.edgeFile = db.barBackdrop.edgeFile
		end
		if not db.castbarBackdrop.insets then
			db.castbarBackdrop.insets = db.barBackdrop.insets
		end
		if profile then
			db.nameFont.name = profile.nameFont and profile.nameFont.name or nil
			db.numFont.name = profile.numFont and profile.numFont.name or nil
			db.castFont.name = profile.castFont and profile.castFont.name or nil
		end
		self:UpdateFonts()
		if doFilterUpdate then
			self:UpdateFilter()
		end
		self:RestyleNamePlates()
		--self:OnTargetChanged()
		--if not self.loading then
		--	self:SetTargetGlow()
		--end
	end

	local function GetJournalLink(instanceID)
		local _, _, _, _, _, _, link, link2 = EJ_GetInstanceInfo(instanceID)
		if type(link2) == "string" then
			link = link2
		end
		if not link then
			return ("[%s %s (%d)]"):format(_G.UNKNOWN, _G.ZONE, instanceID)
		end
		local difficultyID = select(3, GetInstanceInfo())
		if difficultyID == 0 then
			difficultyID = IsInRaid() and GetRaidDifficultyID() or GetDungeonDifficultyID()
		end
		return link:gsub("(Hjournal:%d+:%d+:)(%d+)", "%1"..difficultyID)
	end

	local function GetDiffcultyString()
		local id = select(3, GetInstanceInfo())
		if id == 2 or id == 5 or id == 6 or id == 11 or id == 15 then
			return "Heroic"
		elseif id == 7 or id == 17 then
			return "*" -- LFR
		elseif id == 8 then
			return "Challenge"
		elseif id == 16 or id == 23 then
			return "Mythic"
		elseif id == 24 then
			return "Heroic" -- Timewalker
		end
		return "*" -- Normal
	end

	local diffParents = {
		["Mythic"] = "Heroic",
		["Challenge"] = "Heroic",
		["Heroic"] = "*",
		--["Normal"] = "*",
		--["LFR"] = "*",
		--["Timewalker"] = "Heroic",
	}

	function addon:LoadFilter(zoneFilter, difficulty)
		local hide, recolor, delete, name, color, filter
		local result = {hide = {}, color = {}, zone = zoneFilter.zone}
		while difficulty do
			filter = zoneFilter[difficulty]
			if not filter and difficulty == "*" then
				filter = zoneFilter
			end
			if filter then
				hide = filter.hide
				if type(hide) == "table" then
					for i = 1, #hide do
						name = hide[i]
						if type(name) == "string" and name:len() > 2 then
							delete = name:sub(1, 1) == "-"
							if delete then
								name = name:sub(2)
							end
							if result.hide[name] == nil then
								result.hide[name] = not delete
							end
						end
					end
				end
				recolor = filter.recolor
				if type(recolor) == "table" then
					for k, v in pairs(recolor) do
						--name = type(k) == "number" and (k < 0 and EJ_GetEncounterInfo(-k) or EJ_GetSectionInfo(k)) or k
						if type(k) == "number" and k > 0 then
							name = k
						elseif type(k) == "string" then
							if tonumber(k) then
								name = tonumber(k) < 0 and EJ_GetEncounterInfo(- tonumber(k)) or EJ_GetSectionInfo(tonumber(k))
							else
								name = k
							end
						else
							name = nil
						end
						if name then
							color = (type(v) == "string" and ClassColors[v]) or (type(v) == "number" and db.filterColors[v]) or v
							if result.color[name] == nil then
								if type(name) == "number" then
									result.color[name] = type(color) == "table" and color or false
								else
									result.color[name] = type(color) == "table" and color or false
								end
							end
						end
					end
				end
				if filter.noInherit then break end
			end
			difficulty = diffParents[difficulty]
		end

		for k, v in pairs(result.hide) do if not v then result.hide[k] = nil end end
		for k, v in pairs(result.color) do if type(v) ~= "table" then result.color[k] = nil end end

		return result
	end

	function addon:LoadUserFilter(mapId, baseFilter, rawColor)
		local userFilter = _G.NamePlateKAIFilter and _G.NamePlateKAIFilter[mapId] or nil
		if not userFilter then return end

		local name, color
		for k, v in pairs(userFilter.recolor or {}) do
			if type(k) == "number" and k > 0 then
				name = k
			elseif type(k) == "string" then
				if tonumber(k) then
					name = tonumber(k) < 0 and EJ_GetEncounterInfo(- tonumber(k)) or EJ_GetSectionInfo(tonumber(k))
				else
					name = k
				end
			else
				name = nil
			end
			if name then
				if rawColor then
					baseFilter.color[name] = v
				else
					color = (type(v) == "string" and ClassColors[v]) or (type(v) == "number" and db.filterColors[v]) or v
					if color == false then
						baseFilter.color[name] = nil
					elseif type(color) == "table" then
						baseFilter.color[name] = color
					end
				end
			end
		end
	end

	local cache = {}
	function addon:UpdateFilter()
		--wipe(self.filter.hide)
		wipe(self.filter.color)

		SetMapToCurrentZone()
		local mapId = GetCurrentMapAreaID()
		local difficulty = GetDiffcultyString()
		local zoneFilter = self.defaultFilters[mapId]
		if not zoneFilter then
			mapId = "default"
			zoneFilter = self.defaultFilters.default
		end

		local filter = cache[mapId] and cache[mapId][difficulty] or nil
		if not filter then
			if db.disableDefaultFilters then
				filter = { zone = zoneFilter.zone, color = {}, hide = {} }
			else
				filter = self:LoadFilter(zoneFilter, difficulty)
			end
			self:LoadUserFilter(mapId, filter)
			cache[mapId] = cache[mapId] or {}
			cache[mapId][difficulty] = filter
			local zone = type(filter.zone) == "number" and GetJournalLink(filter.zone) or ("["..(filter.zone or mapId).."]")
			DEFAULT_CHAT_FRAME:AddMessage("|cff3399ff[NamePlateKAI "..addon.version.."]:|r filter loaded "..zone)
		end

		--tShallowCopy(filter.hide, self.filter.hide)
		tShallowCopy(filter.color, self.filter.color)

	end

	function addon:ClearFilterCacheAll()
		for mapId, _ in pairs(cache) do
			self:ClearFilterCache(mapId)
		end
		wipe(cache)
	end

	function addon:ClearFilterCache(mapId)
		if not cache[mapId] then return end
		for difficulty, filter in pairs(cache[mapId]) do
			if type(filter) == "table" then
				if filter.color then
					wipe(filter.color)
				end
				if filter.hide then
					wipe(filter.hide)
				end
				wipe(filter)
			end
		end
		wipe(cache[mapId])
		cache[mapId] = nil
	end
end

do
	addon.frame.highlight = addon.frame:CreateTexture()
	addon.frame.highlight:SetTexture("Interface\\Addons\\NamePlateKAI\\texture\\highlight")
	addon.frame.highlight:SetVertexColor(1, 1, 0.7)
	addon.frame.highlight:SetDrawLayer("OVERLAY")
	addon.frame.highlight:SetAlpha(0.1)
	addon.frame.highlight:SetBlendMode("ADD")

	addon.frame.highlightBG = addon.frame:CreateTexture()
	addon.frame.highlightBG:SetTexture("Interface\\Addons\\NamePlateKAI\\texture\\highlight")
	addon.frame.highlightBG:SetVertexColor(0, 0, 0)
	addon.frame.highlightBG:SetDrawLayer("BACKGROUND")
	addon.frame.highlightBG:SetAlpha(0.6)

	local IsMouseButtonDown = IsMouseButtonDown

	local npcTimer = 0
	local UnitFrame
	function addon.frame:OnUpdate(elapsed) -- self is addon.frame
		if self.taskEnabled then
			addon.scheduler:Update(elapsed)
		end
		if self.highlightFrame and not IsMouseButtonDown(1, 2)and not UnitExists("mouseover") then
			addon:NamePlateOnLeave(self.highlightFrame)
			self.highlight:Hide()
			self.highlightBG:Hide()
			self.highlightFrame = nil
		end
		if npcTimer > 0.15 then
			npcTimer = 0
			UnitFrame = next(npcUnitFrames)
			if UnitFrame then
				npcUnitFrames[UnitFrame] = nil
				UpdateStatusText(UnitFrame, true)
			end
		else
			npcTimer = npcTimer + elapsed
		end
		if #addedPlates > 0 then
			for i = 1, #addedPlates do
				addedPlates[i].UnitFrame:SetAlpha(1)
			end
			wipe(addedPlates)
		end
		if addon.targetPlate then
			addon:UpdateThreatBorder(addon.targetPlate.UnitFrame)
		end
	end
end

-- Target Emphasize
do
	local EmphasizeFrame = CreateFrame("Frame", nil, addon.frame)
	addon.EmphasizeFrame = EmphasizeFrame
	--EmphasizeFrame:SetFrameStrata("BACKGROUND")
	EmphasizeFrame:Hide()

	local shine = EmphasizeFrame:CreateTexture(nil, "BACKGROUND", nil, -7)
	shine:SetTexture("Interface\\Addons\\NamePlateKAI\\texture\\targetglow")
	shine:SetVertexColor(0.1, 0.55, 1, 1)
	shine:SetTexCoord(0, 1, 0, 0.85)

	local shineShadow = EmphasizeFrame:CreateTexture(nil, "BACKGROUND", nil, -8)
	shineShadow:SetTexture("Interface\\Addons\\NamePlateKAI\\texture\\targetbg2")
	shineShadow:SetVertexColor(0, 0, 0, 1)
	shineShadow:SetTexCoord(0, 1, 0, 0.9)
	shineShadow:SetPoint("TOPLEFT", shine, -1, 10)
	shineShadow:SetPoint("BOTTOMRIGHT", shine, 1, 3)

	local shineOverlay = EmphasizeFrame:CreateTexture(nil, "OVERLAY", 6)
	shineOverlay:SetTexture("Interface\\Addons\\NamePlateKAI\\texture\\targetglow")
	shineOverlay:SetVertexColor(0.5, 0.5, 0.5, 1)
	shineOverlay:SetBlendMode("ADD")
	--shineOverlay:SetParent(addon.frame)
	shineOverlay:SetAllPoints(shine)
	--shineOverlay:SetPoint("TOPLEFT", shine, 0, 0)
	--shineOverlay:SetPoint("BOTTOMRIGHT", shine, "RIGHT", 0, 0)

	local flashHighlight = EmphasizeFrame:CreateTexture(nil, "OVERLAY", 7)
	flashHighlight:SetTexture("Interface\\Addons\\NamePlateKAI\\texture\\targetbg2")
	flashHighlight:SetVertexColor(1, 1, 0.6)
	flashHighlight:SetBlendMode("ADD")
	--flashHighlight:SetAllPoints(shineShadow)
	flashHighlight:Hide()

	local targetAnim = EmphasizeFrame:CreateTexture(nil, "OVERLAY")
	targetAnim:SetTexture("Interface\\Addons\\NamePlateKAI\\texture\\targetanimation")
	targetAnim:SetTexCoord(0, 0, 0, 16, 104, 0, 104, 16)
	targetAnim:SetVertexColor(0.5, 0.5, 0.5)
	targetAnim:SetBlendMode("ADD")
	targetAnim:Hide()

	local FLASH_DUR = 0.325
	local FLASH_ALPHA = 0.85

	local AnimateTexCoords, GetTime = AnimateTexCoords, GetTime
	local t
	local function OnUpdate(self, elapsed)
		if db.target.animation then
			AnimateTexCoords(targetAnim, 256, 256, 104, 16, 32, elapsed, 0.005)
		end
		if EmphasizeFrame.fadeIn then
			EmphasizeFrame.fadeIn = false
			EmphasizeFrame:SetAlpha(db.target.shineAlpha)
			EmphasizeFrame.startTime = GetTime()
		elseif EmphasizeFrame.startTime then
			t = FLASH_DUR + EmphasizeFrame.startTime - GetTime()
			if t <= 0 then
				EmphasizeFrame.startTime = nil
				flashHighlight:Hide()
			else
				targetAnim:SetAlpha(db.target.animationAlpha)
				if db.target.flash then
					flashHighlight:SetAlpha((t / FLASH_DUR) * FLASH_ALPHA)
				end
			end
		end
	end
	EmphasizeFrame:SetScript("OnUpdate", OnUpdate)

	function addon:SetTargetGlow(namePlate)
		namePlate = namePlate or self.targetPlate
		if not namePlate then
			--shineShadow:Hide()
			shineOverlay:Hide()
			shineOverlay:SetParent(EmphasizeFrame)
			flashHighlight:Hide()
			flashHighlight:SetParent(EmphasizeFrame)
			targetAnim:Hide()
			targetAnim:SetParent(EmphasizeFrame)
			EmphasizeFrame:Hide()
			EmphasizeFrame:SetParent(addon.frame)
		else
			local bgFrame = namePlate.UnitFrame
			local overlayFrame = namePlate.UnitFrame.RaidTargetFrame
			local width = self.frameWidth
			EmphasizeFrame:SetParent(namePlate.UnitFrame)
			--EmphasizeFrame:SetFrameLevel(1000)
			EmphasizeFrame:ClearAllPoints()
			EmphasizeFrame:SetPoint("BOTTOM", namePlate, "BOTTOM", 0, 0)
			EmphasizeFrame:SetSize(self.frameWidth, self.frameHeight)
			if db.target.animationOverlay then
				targetAnim:SetParent(overlayFrame)
			else
				targetAnim:SetParent(namePlate.UnitFrame.healthBar)
			end
			targetAnim:ClearAllPoints()
			if db.target.animationStyle == "full" then
				targetAnim:SetPoint("BOTTOM", EmphasizeFrame, "BOTTOM", 0, 4 + db.castBarHeight)
				targetAnim:SetPoint("TOP", EmphasizeFrame, "BOTTOM", 0, 11 + db.castBarHeight + db.healthBarHeight + db.nameFont.size + 1)
				targetAnim:SetWidth(width + 16)
			elseif db.target.animationStyle == "wide" then
				targetAnim:SetPoint("BOTTOM", EmphasizeFrame, "BOTTOM", 0, 5 + db.castBarHeight)
				targetAnim:SetPoint("TOP", EmphasizeFrame, "BOTTOM", 0, 11 + db.castBarHeight + db.healthBarHeight)
				targetAnim:SetWidth(width + 16)
			elseif db.target.animationStyle == "fit" then
				targetAnim:SetPoint("BOTTOM", EmphasizeFrame, "BOTTOM", 0, 5 + db.castBarHeight)
				targetAnim:SetPoint("TOP", EmphasizeFrame, "BOTTOM", 0, 11 + db.castBarHeight + db.healthBarHeight)
				targetAnim:SetWidth(width - 16)
			else
				targetAnim:SetPoint("BOTTOM", EmphasizeFrame, "BOTTOM", 0, 8 + db.castBarHeight)
				targetAnim:SetPoint("TOP", EmphasizeFrame, "BOTTOM", 0, 8 + db.castBarHeight + db.healthBarHeight)
				targetAnim:SetWidth(width - 24)
			end
			targetAnim:SetAlpha(db.target.animationAlpha)
			targetAnim:SetShown(db.target.animation)
			shine:ClearAllPoints()
			shine:SetPoint("TOP", EmphasizeFrame, "BOTTOM", 0, 8 + db.castBarHeight + db.healthBarHeight + db.nameFont.size + 4)
			--shine:SetPoint("TOP", namePlate.UnitFrame.name, "TOP", 0, 4)
			shine:SetWidth(width + 34)
			if UnitIsFriend("player", "target") then
				shine:SetVertexColor(0.5, 1, 0.8, 1)
			else
				shine:SetVertexColor(0.6, 0.85, 1, 1)
			end
			shine:SetShown(db.target.shine)
			--shineShadow:SetParent(bgFrame)
			--shineShadow:ClearAllPoints()
			--shineShadow:SetPoint("TOPLEFT", shine, -1, 10)
			--shineShadow:SetPoint("BOTTOMRIGHT", shine, 1, 3)
			shineShadow:SetShown(db.target.shine)
			shineOverlay:SetParent(overlayFrame)
			--shineOverlay:SetAllPoints(shine)
			shineOverlay:SetAlpha(db.target.shineAlpha * 0.5 + 0.5)
			shineOverlay:SetShown(db.target.shine)
			flashHighlight:SetParent(overlayFrame)
			flashHighlight:ClearAllPoints()
			flashHighlight:SetPoint("TOP", EmphasizeFrame, "BOTTOM", 0, 8 + db.castBarHeight + db.healthBarHeight + db.nameFont.size + 14)
			flashHighlight:SetSize(width + 40, 44)
			flashHighlight:SetAlpha(FLASH_ALPHA)
			flashHighlight:SetShown(db.target.flash)
			--EmphasizeFrame.startTime = GetTime()
			EmphasizeFrame:SetAlpha(0)
			EmphasizeFrame:Show()
			EmphasizeFrame.fadeIn = true
		end
	end
end

local function EnableUnitFaction()
	addon.frame:RegisterEvent("UNIT_FACTION")
end

local function DisableUnitFaction()
	addon.frame:UnregisterEvent("UNIT_FACTION")
end

local events = addon.events

function events:PLAYER_TARGET_CHANGED()
	self:OnTargetChanged()
	events.UPDATE_MOUSEOVER_UNIT(self)
end

function events:NAME_PLATE_CREATED(namePlate)
	self:NamePlateOnCreate(namePlate)
end

function events:NAME_PLATE_UNIT_ADDED(unitID)
	self:NamePlateOnAdded(GetNamePlateForUnit(unitID), unitID)
end

function events:NAME_PLATE_UNIT_REMOVED(unitID)
	self:NamePlateOnRemoved(GetNamePlateForUnit(unitID), unitID)
end

function events:UPDATE_MOUSEOVER_UNIT()
	local namePlate = GetNamePlateForUnit("mouseover")
	if namePlate then
		if self.frame.highlightFrame and self.frame.highlightFrame:IsShown() then
			self:NamePlateOnLeave(self.frame.highlightFrame)
		end
		self:NamePlateOnEnter(namePlate)
		self.frame.highlightFrame = namePlate
	end
end

function events:UNIT_NAME_UPDATE(unit)
	local namePlate = GetNamePlateForUnit(unit)
	if namePlate then
		self:RefreshNamePlate(namePlate)
	end
end

function events:UNIT_FACTION(unit)
	if unit == "player" then
		self:UpdatePlayerFaction()
		self.scheduler:Schedule(0.25, "faction", "RefreshAll")
	else
		local namePlate = GetNamePlateForUnit(unit)
		if namePlate then
			self:RefreshNamePlate(namePlate)
		end
	end
end
--events.UNIT_CLASSIFICATION_CHANGED = events.UNIT_FACTION

function events:PLAYER_CONTROL_LOST()
	self.scheduler:Schedule(0.25, "playerControl", "RefreshAll")
end

function events:UNIT_THREAT_SITUATION_UPDATE(unit)
	self:ThreatUpdateAll()
end
events.units.UNIT_THREAT_SITUATION_UPDATE = "player"

function events:ADDON_LOADED(arg)
	if arg == ADDON_NAME then
		self.loading = true
		--DEFAULT_CHAT_FRAME:AddMessage("|cff3399ffNamePlate|r|cff66aaffKAI|r ".."r"..self.version)
		self.targetKai = false
		self.needTapCheck = true

		SystemFont_LargeNamePlate:SetFontObject(FontUnitNameNormal)
		SystemFont_NamePlate:SetFontObject(FontUnitNameNormal)
		SystemFont_NamePlateCastBar:SetFontObject(FontSpellName)

		local profile = _G.NamePlateKAIDB and _G.NamePlateKAIDB.profiles and _G.NamePlateKAIDB.profiles.Default
		if profile then
			profile.version = profile.version or 0
			if profile.version < 7 then -- WoD
				wipe(profile)
			end
			profile.version = 7
		end
	end
end

function addon:UpdatePlayerFaction()
	self.playerFaction = UnitFactionGroup("player") or _G.NONE
end

function events:PLAYER_LOGIN(...)
	hooksecurefunc(NamePlateDriverFrame, "UpdateNamePlateOptions", UpdateNamePlateOptions_Handler)

	self:UpdatePlayerFaction()
	self:LoadProfile()

	if db.healthText then

	end

	if _G.DBM then
		local function DBM_OnEngage(_, mod)
			if type(mod.encounterId) == "number" then
				self:EncounterStart(mod.encounterId)
			end
		end

		local function DBM_OnDisengage(_, mod)
			if type(mod.encounterId) == "number" then
				self:EncounterEnd(mod.encounterId)
			end
		end

		_G.DBM:RegisterCallback("pull", DBM_OnEngage)
		_G.DBM:RegisterCallback("wipe", DBM_OnDisengage)
		_G.DBM:RegisterCallback("kill", DBM_OnDisengage)
	end

	if not self.loading then
		self:UpdateFilter()
	end
	--self:OnTargetChanged()

	DisableUnitFaction()
end

function events:PLAYER_ENTERING_WORLD(...)
	if self.loading then
		events.ZONE_CHANGED_NEW_AREA(self)
	end
	self.loading = nil
	self.frame:SetScript("OnUpdate", addon.frame.OnUpdate)
	self.frame.highlight:Hide()
	self.frame.highlightBG:Hide()
	self.frame.highlightFrame = nil
	if not UnitExists("target") then
		addon:SetTargetGlow()
	end
	self:OnTargetChanged()
	self.scheduler:Schedule(1, "enableUnitFaction", EnableUnitFaction)
end

function events:PLAYER_LEAVING_WORLD(...)
	DisableUnitFaction()
	wipe(npcUnitFrames)
	wipe(addedPlates)
	self:OnTargetChanged(true)
	self.frame:SetScript("OnUpdate", nil)
	self.scheduler:UnscheduleAll()
	self.scheduler:Schedule(1, "encounterEnd", "CheckEncounters")
end

local function InGarrison()
	return C_Garrison.IsPlayerInGarrison(LE_GARRISON_TYPE_6_0) or C_Garrison.IsPlayerInGarrison(LE_GARRISON_TYPE_7_0)
end

local function InBattleZone()
	local instanceName, zoneType = GetInstanceInfo()
	if InGarrison() then
		zoneType = "none"
	end
	if zoneType == "arena" or zoneType == "pvp" then return true, zoneType end
	local zonetext = GetRealZoneText()
	if not zonetext then return false, zoneType end
	for i = 1, GetNumWorldPVPAreas() do -- 1:Wintergrasp 2:Tol Barad 3:Ashran
		local pvpID, localizedName, isActive = GetWorldPVPAreaInfo(i)
		if localizedName == zonetext then return isActive, zoneType end
	end
	return false, zoneType
end

function events:ZONE_CHANGED_NEW_AREA(...)
	self.playerLevel = UnitLevel("player")
	--self.inCombat = UnitAffectingCombat("player")
	local zoneType
	self.inPvP, zoneType = InBattleZone()
	self.needTapCheck = (zoneType == "none") or nil -- if in instance, dont check the tapped
	if zoneType ~= "none" or self.zoneType ~= zoneType then
		self:UpdateFilter()
		self:RefreshAll()
	end
	self.zoneType = zoneType
end

function events:PLAYER_REGEN_ENABLED()
	if self.needRestyle then
		self.needRestyle = false
		self:RestyleNamePlates()
	end
end

function events:PLAYER_REGEN_DISABLED()

end

local function CheckEncounterStart(id)
	if IsEncounterInProgress() then
		addon:EncounterStart(id)
	end
end

function events:ENCOUNTER_START(id, name, ...)
	self.scheduler:Schedule(0.1, "encounterStart", CheckEncounterStart, id)
end

function events:ENCOUNTER_END(id, ...)
	self.scheduler:Schedule(1, "encounterEnd", "CheckEncounters")
end

function events:UNIT_LEVEL(unit)
	if unit == "player" then
		self.playerLevel = UnitLevel("player")
		self.scheduler:Schedule(0.25, "level", "RefreshAll")
	end
end
events.units.UNIT_LEVEL = "player"

function events:BATTLEFIELD_MGR_STATE_CHANGE(id, status)
	if status == 0 or status == 3 then
		self.inPvP = false
	end
end

function events:BATTLEFIELD_MGR_ENTERED(...)
	self.inPvP = true
end

addon.frame:SetFrameStrata("TOOLTIP")
addon.frame:SetScript("OnEvent", function(self, event, ...) events[event](addon, ...) end)
for event, func in pairs(events) do
	if type(func) == "function" then
		if type(events.units[event]) == "string" then
			addon.frame:RegisterUnitEvent(event, events.units[event])
		else
			addon.frame:RegisterEvent(event)
		end
	end
end

local scheduler = addon.scheduler
local tasks = scheduler.tasks

function scheduler:Schedule(duration, id, func, arg1, arg2, arg3)
	local task = id and self:GetTask(id) or nil
	if not task then
		task = {}
		tinsert(tasks, task)
		--lengthTasks = #tasks
		self:UpdateStatus()
	end
	task.r = duration
	task.f = func
	task.id = id
	task.arg1 = arg1
	task.arg2 = arg2
	task.arg3 = arg3
end

function scheduler:GetTask(id)
	for i = 1, #tasks do
		if tasks[i].id == id then
			return tasks[i]
		end
	end
end

function scheduler:UnscheduleAll()
	for i = 1, #tasks do
		wipe(tasks[i])
	end
	wipe(tasks)
	--lengthTasks = 0
	self:UpdateStatus()
end

function scheduler:UpdateStatus()
	addon.frame.taskEnabled = #tasks > 0
end

do
	local type = type
	local i, task
	function scheduler:Update(elapsed)
		i = 1
		while i <= #tasks do
			task = tasks[i]
			if task.r <= 0 then
				if type(task.f) == "string" then
					addon[task.f](addon, task.arg1, task.arg2, task.arg3)
				else
					task.f(task.arg1, task.arg2, task.arg3)
				end
				if task.r <= 0 then
					wipe(tremove(tasks, i))
					self:UpdateStatus()
				end
			else
				task.r = task.r - elapsed
				i = i + 1
			end
		end
	end
end

do -- add panel to blizzard interface options
	local addonName, title, notes, loadable, reason = GetAddOnInfo(ADDON_NAME.."_options")
	if addonName and reason == "DEMAND_LOADED" then
		local panel = CreateFrame("Frame")
		addon.optionsPanel = panel
		panel.name = ADDON_NAME
		panel.title = title or panel.name
		panel:SetScript("OnShow", function()
			if not (LibStub and LibStub("AceGUI-3.0", true)) then
				panel:SetInfoText("AceGUI-3.0 was not found. You must install Ace3.")
			elseif not (LibStub and LibStub("AceDB-3.0", true)) then
				panel:SetInfoText("AceDB-3.0 was not found. You must install Ace3.")
			else
				panel:SetScript("OnShow", nil)
				panel.refresh = nil
				panel:SetInfoText(nil)
				local button = LibStub("AceGUI-3.0"):Create("Button")
				button.frame:SetParent(panel)
				button:SetPoint("TOPLEFT", panel, 12, -12)
				button:SetPoint("TOPRIGHT", panel, -12, -12)
				button:SetText(_G.SHOW.." "..panel.title)
				button:SetCallback("OnClick", function(widget)
					local loaded, reason = LoadAddOn(addonName)
					if loaded then
						widget.frame:SetParent(nil)
						LibStub("AceGUI-3.0"):Release(widget)
					else
						print(YELLOW_FONT_COLOR_CODE.._G.ADDON_LOAD_FAILED:format(panel.title, _G["ADDON_"..reason])..FONT_COLOR_CODE_CLOSE)
					end
				end)
				button.frame:Show()
			end
		end)
		function panel:SetInfoText(text)
			if text and not self.infoText then
				self.infoText = self:CreateFontString(nil, nil, "GameFontNormal")
				self.infoText:SetPoint("TOPLEFT", self, 12, -12)
			end
			if self.infoText then
				self.infoText:SetText(text)
			end
		end
		InterfaceOptions_AddCategory(panel)
		panel:Hide()
	end
end

addon.version = GetAddOnMetadata(ADDON_NAME, "Version")
