local ADDON_NAME, options = ...

local addon = _G.NamePlateKAI
if not addon then return end

local mainPanel = addon.optionsPanel
if not mainPanel then return end

local filterPanel = CreateFrame("Frame")
filterPanel.name = _G.FILTERS
filterPanel.parent = mainPanel.name

local function print(...)
	_G.print(YELLOW_FONT_COLOR_CODE..mainPanel.title..":"..FONT_COLOR_CODE_CLOSE, ...)
end

local GUI = LibStub and LibStub("AceGUI-3.0", true)
if not GUI then return print("AceGUI-3.0 was not found.") end
if not (LibStub and LibStub("AceDB-3.0", true)) then return print("AceDB-3.0 was not found.") end
local db = LibStub("AceDB-3.0"):New("NamePlateKAIDB", {profile = addon.defaults}, true)
local profile = db.profile

local FRAME_MARGIN = 14
local FRAME_SPACE = 4

local controls = {}
local filterControls = {}
local zoneFilters = {}
local userFilters = {}

local function Animation_OnValueChanged(widget, event, value)
	local disabled = not controls.animationCheck:GetValue()
	controls.animationAlphaSlider:SetDisabled(disabled)
	controls.animationStyleDropdown:SetDisabled(disabled)
	controls.animationAlphaSlider:SetDisabled(disabled)
	controls.animationOverlayCheck:SetDisabled(disabled)
end

local function Shine_OnValueChanged(widget, event, value)
	local disabled = not controls.shineCheck:GetValue()
	controls.shineAlphaSlider:SetDisabled(disabled)
end

local function FontDropdownSetValue(dropdown, fontPath)
	if fontPath then
		for k, v in pairs(dropdown.list) do
			if v == fontPath then
				dropdown:SetValue(k)
				break
			end
		end
	else
		dropdown:SetValue(_G.DEFAULT)
	end
end

function mainPanel:refresh()
	if not next(controls) then
		options:CreateWidget()
		options:CreateFiltersWidget()
	end

	--controls.healthWidthSlider:SetValue(profile.frameSize.W)
	--controls.healthHeightSlider:SetValue(profile.frameSize.H)
	controls.frameScaleHorizontal:SetValue(profile.frameScale.x)
	--controls.frameScaleVertical:SetValue(profile.frameScale.y)

	controls.healthHeightSlider:SetValue(profile.healthBarHeight)
	controls.castHeightSlider:SetValue(profile.castBarHeight)

	controls.frameMarginY:SetValue(profile.frameMargin.y)
	controls.raidIconSlider:SetValue(profile.raidIcon.size)
	controls.spellIconSlider:SetValue(profile.spellIcon.size)

	controls.unitNameSlider:SetValue(profile.nameFont.size)
	controls.spellNameSlider:SetValue(profile.castFont.size)
	--controls.numberSlider:SetValue(profile.numFont.size)

	if controls.unitFontDropdown then
		FontDropdownSetValue(controls.unitFontDropdown, profile.nameFont.name)
	end

	if controls.spellFontDropdown then
		FontDropdownSetValue(controls.spellFontDropdown, profile.castFont.name)
	end

	controls.flashCheck:SetValue(profile.target.flash)
	controls.shineCheck:SetValue(profile.target.shine)
	controls.shineAlphaSlider:SetValue(profile.target.shineAlpha)
	controls.animationCheck:SetValue(profile.target.animation)
	controls.animationAlphaSlider:SetValue(profile.target.animationAlpha)
	controls.animationStyleDropdown:SetValue(profile.target.animationStyle)
	controls.animationOverlayCheck:SetValue(profile.target.animationOverlay)
	Shine_OnValueChanged()
	Animation_OnValueChanged()

	controls.range40Check:SetValue(GetCVar("nameplateMaxDistance") == "40")
	controls.clampTopCheck:SetValue(GetCVar("nameplateOtherTopInset") ~= "-1")
	controls.clampBottomCheck:SetValue(GetCVar("nameplateOtherBottomInset") ~= "-1")
	controls.targetAlwaysOnTop:SetValue(profile.targetAlwaysOnTop)
	controls.redHostileCheck:SetValue(profile.redHostile)
	controls.disableFactionIconCheck:SetValue(profile.disableFactionIcon)
	controls.friendlyClassColorCheck:SetValue(profile.friendlyClassColor)

	controls.disableDefaultFiltersCheck:SetValue(profile.disableDefaultFilters)

	--controls.reloadButton:SetDisabled(true)
end

function mainPanel:okay()
	--profile.frameSize.W = controls.healthWidthSlider:GetValue()
	--profile.frameSize.H = controls.healthHeightSlider:GetValue()
	profile.frameScale.x = controls.frameScaleHorizontal:GetValue()
	--profile.frameScale.y = controls.frameScaleVertical:GetValue()
	profile.healthBarHeight = controls.healthHeightSlider:GetValue()
	profile.castBarHeight = controls.castHeightSlider:GetValue()

	profile.frameMargin.y = controls.frameMarginY:GetValue()
	profile.raidIcon.size = controls.raidIconSlider:GetValue()
	profile.spellIcon.size = controls.spellIconSlider:GetValue()

	profile.nameFont.size = controls.unitNameSlider:GetValue()
	profile.castFont.size = controls.spellNameSlider:GetValue()
	--profile.numFont.size = controls.numberSlider:GetValue()

	local fontName = controls.unitFontDropdown and controls.unitFontDropdown:GetValue()
	if fontName then
		profile.nameFont.name = (fontName ~= _G.DEFAULT) and controls.unitFontDropdown.list[fontName] or nil
	end

	local fontName = controls.spellFontDropdown and controls.spellFontDropdown:GetValue()
	if fontName then
		profile.castFont.name = (fontName ~= _G.DEFAULT) and controls.spellFontDropdown.list[fontName] or nil
	end

	profile.target.flash = controls.flashCheck:GetValue()
	profile.target.shine = controls.shineCheck:GetValue()
	profile.target.shineAlpha = controls.shineAlphaSlider:GetValue()
	profile.target.animation = controls.animationCheck:GetValue()
	profile.target.animationAlpha = controls.animationAlphaSlider:GetValue()
	profile.target.animationStyle = controls.animationStyleDropdown:GetValue()
	profile.target.animationOverlay = controls.animationOverlayCheck:GetValue()

	local value
	value = controls.range40Check:GetValue()
	if value == (GetCVar("nameplateMaxDistance") ~= "40") then
		addon:QueueSetCVar("nameplateMaxDistance", value and 40 or GetCVarDefault("nameplateMaxDistance"))
	end

	value = controls.clampTopCheck:GetValue()
	if value == (GetCVar("nameplateOtherTopInset") == "-1") then
		addon:QueueSetCVar("nameplateOtherTopInset", value and GetCVarDefault("nameplateOtherTopInset") or -1)
	end

	value = controls.clampBottomCheck:GetValue()
	if value == (GetCVar("nameplateOtherBottomInset") == "-1") then
		addon:QueueSetCVar("nameplateOtherBottomInset", value and GetCVarDefault("nameplateOtherBottomInset") or -1)
	end

	profile.targetAlwaysOnTop = controls.targetAlwaysOnTop:GetValue()
	profile.redHostile = controls.redHostileCheck:GetValue()
	profile.disableFactionIcon = controls.disableFactionIconCheck:GetValue()
	profile.friendlyClassColor = controls.friendlyClassColorCheck:GetValue()

	local value = controls.disableDefaultFiltersCheck:GetValue() and true
	if profile.disableDefaultFilters ~= value then
		profile.disableDefaultFilters = value
		addon:ClearFilterCacheAll()
	end

	options:SaveUserFilters()
	options:ClearUserFilters()
	addon:LoadProfile(true)
end

function mainPanel:default()
	db:ResetProfile()
	mainPanel:refresh()
	mainPanel:okay()
end

function filterPanel:refresh()
	options:RefreshFilters()
end

function filterPanel:okay()

end

function filterPanel:cancel()
	options:ClearUserFilters()
end

function filterPanel:default()
	HideUIPanel(ColorPickerFrame)
	options:ResetUserFilters()
	options:ClearUserFilters()
	addon:UpdateFilter()
	addon:RestyleNamePlates()
	filterPanel:refresh()
end

local function NOP() end

local function Slider_OnMouseWheel(frame, delta)
	local widget = frame.obj
	if not widget.disabled then
		widget:SetValue(math.min(math.max(widget.value + (widget.step * delta), widget.min), widget.max))
	end
end

local function CreateGroup(name, title, parent)
	local group = GUI:Create("InlineGroup")
	group.frame:SetParent(parent or mainPanel)
	group:SetTitle(title)
	group:SetLayout("Flow")
	group.frame:EnableMouseWheel(true)
	group.frame:SetScript("OnMouseWheel", NOP)
	--group.frame:Show()
	controls[name] = group
	return group
end

local function CreateSlider(name, label, min, max, step, width, onValueChanged)
	if not onValueChanged and type(width) == "function" then
		onValueChanged = width
		width = nil
	end
	local slider = GUI:Create("Slider")
	slider:SetWidth(width or 142)
	slider:SetLabel(label)
	slider:SetSliderValues(min, max, step)
	slider.frame:EnableMouseWheel(true)
	slider.frame:SetScript("OnMouseWheel", Slider_OnMouseWheel)
	if type(onValueChanged) == "function" then
		slider:SetCallback("OnValueChanged", onValueChanged)
	end
	controls[name] = slider
	return slider
end

local function CreateHeader(name)
	local header = GUI:Create("Heading")
	header:SetFullWidth(true)
	if name then
		controls[name] = header
	end
	return header
end

local function CreateCheckBox(name, label, description, width, onValueChanged)
	local check = GUI:Create("CheckBox")
	if width > 0 then
		check:SetRelativeWidth(width)
	else
		check:SetWidth(-width)
	end
	check:SetLabel(label)
	if description then
		check:SetDescription(description)
	end
	if type(onValueChanged) == "function" then
		check:SetCallback("OnValueChanged", onValueChanged)
	end
	controls[name] = check
	return check
end

--local function Slider_OnValueChanged(widget, event)
--	controls.reloadButton:SetDisabled(false)
--end

--local function chatBubbleCheck_OnValueChanged(widget, event, value)
--	controls.chatBubbleSlider:SetDisabled(not value)
--	controls.reloadButton:SetDisabled(false)
--end

function options:CreateWidget()
	local titleText = mainPanel:CreateFontString(nil, nil, "GameFontNormalLarge")
	titleText:SetText(mainPanel.name)
	titleText:SetPoint("TOPLEFT", mainPanel, FRAME_MARGIN, -FRAME_MARGIN)

	local group = CreateGroup("styleGroup", "Style")
	group:SetPoint("TOPLEFT", mainPanel, FRAME_MARGIN, -36)
	group:SetPoint("TOPRIGHT", mainPanel, -FRAME_MARGIN, -36)
	group.frame:Show()

	--group:AddChild(CreateSlider("healthWidthSlider", "Frame Width", 50, 200, 1))
	--group:AddChild(CreateSlider("healthHeightSlider", "Height (Health)", 3, 30, 0.1))
	group:AddChild(CreateSlider("frameScaleHorizontal", "Frame Width", 0.5, 2.0, 0.01))
	--group:AddChild(CreateSlider("frameScaleVertical", "Scale Y", 0.5, 5.0, 0.05))
	group:AddChild(CreateSlider("frameMarginY", "Frame Margin", 0, 60, 1))
	group:AddChild(CreateSlider("healthHeightSlider", "Healthbar Height", 3, 25, 0.1))
	group:AddChild(CreateSlider("castHeightSlider", "Castbar Height", 3, 25, 0.1))
	--group:AddChild(CreateHeader())
	group:AddChild(CreateSlider("unitNameSlider", "Unit Font Size", 5, 30, 0.1))
	group:AddChild(CreateSlider("spellNameSlider", "Spell Font Size", 5, 30, 0.1))
	group:AddChild(CreateSlider("raidIconSlider", "Raid Icon Size", 5, 40, 1))
	--group:AddChild(CreateSlider("numberSlider", "Number Size", 5, 30, 0.1))
	group:AddChild(CreateSlider("spellIconSlider", "Spell Icon Size", 5, 30, 0.1))

	if AceGUIWidgetLSMlists and AceGUIWidgetLSMlists.font then
		local unitFontDropdown = GUI:Create("LSM30_Font")
		unitFontDropdown:SetLabel("Unit Font")
		unitFontDropdown:SetRelativeWidth(0.35)
		local unitFontList = {[_G.DEFAULT] = _G.GameFontWhite:GetFont()}
		for k, v in pairs(AceGUIWidgetLSMlists.font) do
			unitFontList[k] = v
		end
		unitFontDropdown:SetList(unitFontList)
		unitFontDropdown:SetCallback("OnValueChanged", function(widget, event, key)
			widget:SetValue(key)
		end)
		controls.unitFontDropdown = unitFontDropdown
		group:AddChild(unitFontDropdown)

		local spellFontDropdown = GUI:Create("LSM30_Font")
		spellFontDropdown:SetLabel("Spell Font")
		spellFontDropdown:SetRelativeWidth(0.35)
		local spellFontList = {[_G.DEFAULT] = _G.GameFontWhiteTiny:GetFont()}
		for k, v in pairs(AceGUIWidgetLSMlists.font) do
			spellFontList[k] = v
		end
		spellFontDropdown:SetList(spellFontList)
		spellFontDropdown:SetCallback("OnValueChanged", function(widget, event, key)
			widget:SetValue(key)
		end)
		controls.spellFontDropdown = spellFontDropdown
		group:AddChild(spellFontDropdown)
	end

	local group = CreateGroup("targetGroup", "Target Emphasize")
	group:SetPoint("TOPLEFT", controls.styleGroup.frame, "BOTTOMLEFT", 0, -FRAME_SPACE)
	group:SetPoint("TOPRIGHT", controls.styleGroup.frame, "BOTTOMRIGHT", 0, -FRAME_SPACE)
	group:AddChild(CreateCheckBox("flashCheck", "Flash", nil, 0.15))
	group:AddChild(CreateCheckBox("shineCheck", "Shine", nil, 0.15, Shine_OnValueChanged))
	group:AddChild(CreateSlider("shineAlphaSlider", "Shine Alpha", 0.1, 1.0, 0.05))
	group:AddChild(CreateHeader())
	group:AddChild(CreateCheckBox("animationCheck", "Animation", nil, 0.2, Animation_OnValueChanged))
	group:AddChild(CreateSlider("animationAlphaSlider", "Animation Alpha", 0.01, 1.0, 0.01))

	local animationStyleDropdown = GUI:Create("Dropdown")
	animationStyleDropdown:SetLabel("Animation Style")
	animationStyleDropdown:SetRelativeWidth(0.2)
	animationStyleDropdown:SetList({default = _G.DEFAULT, fit = "Fit", wide = "Wide", full = "Full"}, {"default", "fit", "wide", "full"})
	controls.animationStyleDropdown = animationStyleDropdown
	group:AddChild(animationStyleDropdown)

	group:AddChild(CreateCheckBox("animationOverlayCheck", "Overlay", nil, 0.15))
	group.frame:Show()

	local group = CreateGroup("commonGroup", "Common")
	group:SetPoint("TOPLEFT", controls.targetGroup.frame, "BOTTOMLEFT", 0, -FRAME_SPACE)
	group:SetPoint("TOPRIGHT", controls.targetGroup.frame, "BOTTOMRIGHT", 0, -FRAME_SPACE)

	group:AddChild(CreateCheckBox("range40Check", "Show within 40 yd", nil, 0.27))
	group:AddChild(CreateCheckBox("clampTopCheck", "Clamp to the screen top", nil, 0.33))
	group:AddChild(CreateCheckBox("clampBottomCheck", "Clamp to the screen bottom", nil, 0.4))
	group:AddChild(CreateCheckBox("targetAlwaysOnTop", "Target is always on top", nil, 0.35))
	group:AddChild(CreateCheckBox("disableFactionIconCheck", "Disable faction icon in PvP", nil, 0.36))
	group:AddChild(CreateCheckBox("redHostileCheck", "Use red color for hostile NPC", nil, 0.4))
	group:AddChild(CreateCheckBox("friendlyClassColorCheck", "Use class colors for friendly players", nil, 0.48))

	group.frame:Show()

	local button = GUI:Create("Button") -- Apply
	button.frame:SetParent(mainPanel)
	button:SetWidth(100)
	button:SetPoint("BOTTOMRIGHT", mainPanel, -FRAME_MARGIN, FRAME_MARGIN)
	button:SetText(_G.APPLY)
	button:SetCallback("OnClick", function()
		mainPanel:okay()
		mainPanel:refresh()
	end)
	button.frame:Show()

--	local button = GUI:Create("Button") -- Reload
--	button.frame:SetParent(mainPanel)
--	button:SetWidth(100)
--	button:SetPoint("BOTTOMLEFT", mainPanel, FRAME_MARGIN, FRAME_MARGIN)
--	button:SetText(_G.RELOADUI)
--	--button:SetDisabled(true)
--	button:SetCallback("OnClick", function()
--		mainPanel:okay()
--		ReloadUI()
--	end)
--	button.frame:Show()
--	controls.reloadButton = button
end

local NpcNames do
	local tooltip = CreateFrame("GameTooltip")
	tooltip.leftLines = {}
	tooltip.rightLines = {}
	for i = 1, 2 do
		local L,R = tooltip:CreateFontString(), tooltip:CreateFontString()
		L:SetFontObject(GameFontNormal)
		R:SetFontObject(GameFontNormal)
		tooltip:AddFontStrings(L,R)
		tooltip.leftLines[i] = L
	end
	local FORMAT_TOOLTIP_LEVEL = "^"..TOOLTIP_UNIT_LEVEL:gsub("%%s", ".+")
	local instanceID = select(8, GetInstanceInfo())
	local FORMAT_GUID_NPC = "unit:Creature-0-1-1-12345-%d-0000000000"
	NpcNames = setmetatable({}, {__index =
		function(self, key)
			tooltip:SetOwner(_G.UIParent, "ANCHOR_NONE")
			tooltip:SetHyperlink(FORMAT_GUID_NPC:format(key))
			local value = tooltip.leftLines[1]:GetText()
			tooltip:Hide()
			if value then
				self[key] = value
			end
			return value or _G.UNKNOWN
		end
	})
end

function options:ResetUserFilters()
	if _G.NamePlateKAIFilter then
		for map, filter in pairs(_G.NamePlateKAIFilter) do
			if filter.recolor then
				for name, color in pairs(filter.recolor) do
					if type(color) == "table" then
						wipe(color)
					end
				end
				wipe(filter.recolor)
			end
			wipe(filter)
			--addon:ClearFilterCache(map)
		end
		wipe(_G.NamePlateKAIFilter)
	end
	addon:ClearFilterCacheAll()
end

function options:SaveUserFilters()
	if not _G.NamePlateKAIFilter then
		_G.NamePlateKAIFilter = {}
	end
	local db = _G.NamePlateKAIFilter

	for map, filter in pairs(userFilters) do
		if filter.modified then
			if not db[map] then db[map] = { recolor = {} } end
			if not db[map].recolor then db[map].recolor = {} end
			wipe(db[map].recolor)
			for name, color in pairs(filter.color) do
				db[map].recolor[name] = color
			end
			if next(db[map].recolor) == nil then
				wipe(db[map])
				db[map] = nil
			end
			addon:ClearFilterCache(map)
			filter.modified = nil
		end
	end
	if not next(db) then
		_G.NamePlateKAIFilter = nil
	end
end

function options:ClearUserFilters()
	for map, filter in pairs(userFilters) do
		if filter.color then
			wipe(filter.color)
		end
		wipe(filter)
	end
	wipe(userFilters)

	controls.filterGroup:ReleaseChildren()
	wipe(filterControls)
end

local function GetColoredText(text, color)
	return ("|cff%.2x%.2x%.2x%s|r"):format(color[1] * 255, color[2] * 255, color[3] * 255, text)
end

local function GetColoredSqure(text, color)
	return ("|T130871:0:0:0:0:8:8:0:8:0:8:%d:%d:%d|t"):format(color[1] * 255, color[2] * 255, color[3] * 255)..(text and GetColoredText(text, color) or "")
end

local function LoadZoneFilter(map)
	if not addon.defaultFilters[map] then return end
	if not zoneFilters[map] then
		local t = {}
		for _, difficulty in pairs({"*", "Heroic", "Challenge", "Mythic"}) do
			if difficulty == "*" or addon.defaultFilters[map][difficulty] then
				local filter = addon:LoadFilter(addon.defaultFilters[map], difficulty)
				for k, v in pairs(filter.color) do
					if not t[k] then t[k] = {} end
					t[k][difficulty] = v
				end
			end
		end
		zoneFilters[map] = t
	end
	return zoneFilters[map]
end

local function RefreshFilterNpcNames()
	for k, v in pairs(filterControls) do
		if k:find("\\nameLabel$") then
			local id = v:GetUserData("id")
			if type(id) == "number" then
				v:SetText(NpcNames[id])
			end
		end
	end
end

local filterOrder = {}
function options:RefreshFilters()
	controls.filterGroup:ReleaseChildren()
	wipe(filterControls)
	local map = controls.filterZoneDropdown:GetValue()
	if not map then
		SetMapToCurrentZone()
		map = GetCurrentMapAreaID() or "default"
		if not controls.filterZoneDropdown.list[map] then
			map = "default"
		end
		controls.filterZoneDropdown:SetValue(map)
	end

	controls.filterGroup:PauseLayout()

	local zoneFilter = LoadZoneFilter(map)

	wipe(filterOrder)
	local order = addon.defaultFilters[map] and addon.defaultFilters[map].order
	if type(order) == "table" then
		local id
		for i, v in ipairs(order) do
			id = type(v) == "string" and tonumber(v)
			if id then
				tinsert(filterOrder, id < 0 and EJ_GetEncounterInfo(-id) or EJ_GetSectionInfo(id) or id)
			else
				tinsert(filterOrder, v)
			end
		end
	end

	local header = GUI:Create("Label")
	header:SetFullWidth(true)
	controls.filterGroup:AddChild(header)

	if not userFilters[map] then
		userFilters[map] = { color = {} }
		addon:LoadUserFilter(map, userFilters[map], true)
	end

	local count = 0

	for name, v in pairs(zoneFilter) do
		if not tContains(filterOrder, name) then
			options:AddFilterWidget(map, name, zoneFilter)
			count = count + 1
		end
	end

	local header
	for _, name in ipairs(filterOrder) do
		headerText = type(name) == "string" and name:match("h\\(.+)")
		if headerText then
			if tonumber(headerText) then
				headerText = EJ_GetEncounterInfo(headerText) or headerText
			end
			local header = GUI:Create("Heading")
			header:SetFullWidth(true)
			if headerText ~= "-" then
				header:SetText(headerText)
			end
			controls.filterGroup:AddChild(header)
		elseif zoneFilter[name] then
			options:AddFilterWidget(map, name, zoneFilter)
			count = count + 1
		end
	end

	--for name, v in pairs(zoneFilter) do
	--	options:AddFilterWidget(map, name, zoneFilter)
	--	count = count + 1
	--end

	if count > 0 then
		local header = GUI:Create("Heading")
		header:SetFullWidth(true)
		controls.filterGroup:AddChild(header)
	end

	for name, v in pairs(userFilters[map].color) do
		if not zoneFilter[name] then
			options:AddFilterWidget(map, name)
			count = count + 1
		end
	end

	if count > 0 then
		C_Timer.After(1, RefreshFilterNpcNames)
	else
		local noneLabel = GUI:Create("Label")
		noneLabel:SetRelativeWidth(0.3)
		noneLabel:SetFontObject(GameFontNormal)
		noneLabel:SetText(_G.NONE)
		controls.filterGroup:AddChild(noneLabel)
		local header = GUI:Create("Heading")
		header:SetFullWidth(true)
		controls.filterGroup:AddChild(header)
	end

	options:AddFilterButton()

	controls.filterGroup:ResumeLayout()
	controls.filterGroup:DoLayout()
end

local function GetZoneList()
	local list = {}
	local order = {}
	local name
	for map, filter in pairs(addon.defaultFilters) do
		if type(filter.zone) == "number" then
			name = EJ_GetInstanceInfo(filter.zone)
		else
			name = filter.zone
		end
		if not name then
			name = _G.UNKNOWN.."["..map.."]"
		end
		if type(filter.category) == "number" then
			if filter.category > 700 then
				name = "|cff44ff44"..name.."|r"
			elseif filter.category > 600 then
				name = "|cffffcc44"..name.."|r"
			elseif filter.category > 500 then
				name = "|cff44ffcc"..name.."|r"
			elseif filter.category > 400 then
				name = "|cffff4444"..name.."|r"
			end
			if (filter.category % 100) > 90 then
				name = name.." (".._G.PVP..")"
			elseif (filter.category % 100) > 50 then
				name = name.." (".._G.RAID..")"
			end
		end
		tinsert(order, map)
		list[map] = name
	end
	local p1, p2
	table.sort(order, function(v1, v2)
		v1 = tonumber(v1) or v1
		v2 = tonumber(v2) or v2
		p1 = tonumber(addon.defaultFilters[v1] and addon.defaultFilters[v1].category or 10000)
		p2 = tonumber(addon.defaultFilters[v2] and addon.defaultFilters[v2].category or 10000)
		if math.floor(p1 / 100) == math.floor(p2 / 100) then
			return p1 < p2
		else
			return math.floor(p1 / 100) > math.floor(p2 / 100)
		end
	end)
	return list, order
end

local function GetFilterDefaultText(zoneFilter, name)
	local text
	local color = zoneFilter[name]["*"]
	if color then
		text = GetColoredSqure(_G.DEFAULT, color)
	else
		text = GetColoredText(_G.DEFAULT, profile.barColors.tapped)
	end
	text = text.."  "
	color = zoneFilter[name]["Heroic"]
	if color then
		text = text..GetColoredSqure("H ", color)
	end
	color = zoneFilter[name]["Challenge"]
	if color then
		text = text..GetColoredSqure("C ", color)
	end
	color = zoneFilter[name]["Mythic"]
	if color then
		text = text..GetColoredSqure("M ", color)
	end
	return text
end

local function FilterColorDropdown_Refresh(dropdown)
	local map = dropdown:GetUserData("map")
	local name = dropdown:GetUserData("id")
	local value = "default"
	local colorPicker = filterControls[name.."\\colorPicker"]
	--if not colorPicker then return end
	if userFilters[map] then
		local color = userFilters[map].color[name]
		if color == false then
			value = "disable"
		elseif type(color) == "number" then
			value = color
		elseif type(color) == "table" then
			value = "custom"
			colorPicker = colorPicker or options:AddFilterColorPicker(map, name)
			colorPicker:SetColor(color[1], color[2], color[3])
		end
	end
	dropdown:SetValue(value)
	if colorPicker then
		colorPicker.frame:SetShown(value == "custom")
	end
end

local filterColors = {
	[1] = GetColoredSqure("1 Light Yellow", profile.filterColors[1]),
	[2] = GetColoredSqure("2 Light Purple", profile.filterColors[2]),
	[3] = GetColoredSqure("3 Light Blue", profile.filterColors[3]),
	[4] = GetColoredSqure("4 (no aggro)", profile.filterColors[4]),
	[5] = GetColoredSqure("5 (no aggro)", profile.filterColors[5]),
	[6] = GetColoredSqure("6 (no aggro)", profile.filterColors[6]),
	disable = _G.DISABLE,
	custom = _G.CUSTOM,
}
local colorOrder = {"default", "disable", 1, 2, 3, 4, 5, 6, "custom"}
local userOrder = {"disable", 1, 2, 3, 4, 5, 6, "custom", "remove"}

--do
--	for name, c in pairs(RAID_CLASS_COLORS) do
--		filterColors[name] = GetColoredSqure(name, {c.r, c.g, c.b})
--		tinsert(colorOrder, name)
--	end
--end

local function FilterColorDropdown_OnValueChanged(dropdown, event, key)
	HideUIPanel(ColorPickerFrame)
	local map = dropdown:GetUserData("map")
	local name = dropdown:GetUserData("id")
	local userFilter = userFilters[map]
	if not userFilter then return end

	if key == "default" then
		userFilter.color[name] = nil
	elseif key == "disable" then
		userFilter.color[name] = false
	elseif key == "custom" then
		if type(userFilter.color[name]) ~= "table" then
			local colorPicker = filterControls[name.."\\colorPicker"]
			if colorPicker then
				userFilter.color[name] = {colorPicker.r, colorPicker.g, colorPicker.b}
			else
				userFilter.color[name] = {unpack(profile.filterColors[1], 1, 3)}
			end
		end
	elseif key == "remove" then
		userFilter.color[name] = nil
		userFilter.modified = true
		if not zoneFilters[map][name] then
			options:RemoveFilterWidget(map, name)
			controls.filterGroup:DoLayout()
			return
		end
	elseif type(key) == "number" then
		userFilter.color[name] = key
	else
		return
	end
	userFilter.modified = true
	FilterColorDropdown_Refresh(dropdown)
end

local function FilterZoneDropdown_OnValueChanged(widget, event, key)
	HideUIPanel(ColorPickerFrame)
	options:RefreshFilters()
end

local function FilterColorPicker_OnValueConfirmed(colorPicker, event, r, g, b, a)
	local map = colorPicker:GetUserData("map")
	local name = colorPicker:GetUserData("id")
	local color = userFilters[map] and userFilters[map].color[name]
	if type(color) == "table" and (color[1] ~= r or color[2] ~= g or color[3] ~= b) then
		wipe(color)
		color[1] = r
		color[2] = g
		color[3] = b
		userFilters[map].modified = true
	end
end

local function FilterResetButton_OnClick(button, event)
	HideUIPanel(ColorPickerFrame)
	local map = controls.filterZoneDropdown:GetValue()
	if not map then return end
	if not userFilters[map] then
		userFilters[map] = { color = {} }
	end
	local userFilter = userFilters[map]
	wipe(userFilter.color)
	userFilter.modified = true
	options:RefreshFilters()
end

function options:CreateFiltersWidget()
	local filterTitleText = filterPanel:CreateFontString(nil, nil, "GameFontNormalLarge")
	filterTitleText:SetText(mainPanel.name.." - "..filterPanel.name)
	filterTitleText:SetPoint("TOPLEFT", filterPanel, FRAME_MARGIN, -FRAME_MARGIN)

	local checkBox = CreateCheckBox("disableDefaultFiltersCheck", "Disable all bult-in filters", nil, -210)
	checkBox.frame:SetParent(filterPanel)
	checkBox:SetPoint("TOPRIGHT", filterPanel, -FRAME_MARGIN, -FRAME_MARGIN)
	checkBox.frame:Show()

	local group = CreateGroup("filterGroup", "", filterPanel)
	group:SetPoint("TOPLEFT", filterPanel, FRAME_MARGIN, -26)
	group:SetPoint("BOTTOMRIGHT", filterPanel, -FRAME_MARGIN, FRAME_MARGIN -4)
	group.frame:Show()

	local filterZoneDropdown = GUI:Create("Dropdown")
	filterZoneDropdown.text:SetJustifyH("CENTER")
	--filterZoneDropdown:SetFullWidth(true)
	filterZoneDropdown:SetRelativeWidth(0.84)
	filterZoneDropdown:SetList(GetZoneList())
	filterZoneDropdown:SetCallback("OnValueChanged", FilterZoneDropdown_OnValueChanged)
	hooksecurefunc(filterZoneDropdown.pullout, "Open", function()
		HideUIPanel(ColorPickerFrame)
	end)
	controls.filterZoneDropdown = filterZoneDropdown
	group:AddChild(filterZoneDropdown)

	local resetButton = GUI:Create("Button") -- Reset
	resetButton:SetRelativeWidth(0.15)
	resetButton:SetText(_G.RESET)
	resetButton:SetCallback("OnClick", FilterResetButton_OnClick)
	controls.resetButton = resetButton
	group:AddChild(resetButton)

	local filterGroup = GUI:Create("ScrollFrame")
	group:AddChild(filterGroup)
	filterGroup:SetLayout("Flow")
	filterGroup:SetFullHeight(true)
	filterGroup:SetFullWidth(true)
	controls.filterGroup = filterGroup
end

function options:AddFilterWidget(map, name, zoneFilter)
	if not name then return end

	local nextWidget = filterControls.nameEdit

	local idLabel = GUI:Create("Label")
	idLabel:SetUserData("id", name)
	idLabel:SetRelativeWidth(0.13)
	idLabel:SetFontObject(GameFontNormal)
	idLabel:SetText(type(name) == "number" and name or nil)
	filterControls[name.."\\idLabel"] = idLabel
	controls.filterGroup:AddChild(idLabel, nextWidget)

	local nameLabel = GUI:Create("Label")
	nameLabel:SetUserData("id", name)
	nameLabel:SetRelativeWidth(0.52)
	nameLabel:SetFontObject(GameFontNormal)
	local text = name
	if type(name) == "number" then
		text = NpcNames[name]
	end
	nameLabel:SetText(text or _G.UNKNOWN)
	filterControls[name.."\\nameLabel"] = nameLabel
	controls.filterGroup:AddChild(nameLabel, nextWidget)

	local colorDropdown = GUI:Create("Dropdown")
	colorDropdown:SetUserData("id", name)
	colorDropdown:SetUserData("map", map)
	colorDropdown:SetRelativeWidth(0.27)
	local text
	local list, order
	if zoneFilter then
		list = {default = GetFilterDefaultText(zoneFilter, name)}
		order = colorOrder
	else
		list = {remove = _G.DELETE}
		order = userOrder
	end
	for k, v in pairs(filterColors) do
		list[k] = v
	end
	colorDropdown:SetList(list, order)
	colorDropdown:SetCallback("OnValueChanged", FilterColorDropdown_OnValueChanged)
	filterControls[name.."\\colorDropdown"] = colorDropdown
	controls.filterGroup:AddChild(colorDropdown, nextWidget)

	local footer = GUI:Create("Label")
	footer:SetUserData("id", name)
	footer:SetFullWidth(true)
	filterControls[name.."\\footer"] = footer
	controls.filterGroup:AddChild(footer, nextWidget)

	--controls.filterGroup:AddChildren(idLabel, nameLabel, colorDropdown, footer)

	FilterColorDropdown_Refresh(colorDropdown)
end

function options:AddFilterColorPicker(map, name)
	local colorPicker = GUI:Create("ColorPicker")
	colorPicker:SetUserData("id", name)
	colorPicker:SetUserData("map", map)
	colorPicker:SetRelativeWidth(0.04)
	colorPicker:SetColor(0, 0, 0)
	colorPicker:SetHasAlpha(false)
	colorPicker:SetCallback("OnValueConfirmed", FilterColorPicker_OnValueConfirmed)
	filterControls[name.."\\colorPicker"] = colorPicker
	controls.filterGroup:AddChild(colorPicker, filterControls[name.."\\footer"])
	return colorPicker
end

function options:AddFilter(map, name)
	if not map then return end
	if not name then return end
	name = tonumber(name) or name

	local zoneFilter = zoneFilters[map]
	if not zoneFilter then return end
	if zoneFilter[name] then
		return "Filter already exists: "..name
	end

	if not userFilters[map] then
		userFilters[map] = { color = {} }
	end
	local userFilter = userFilters[map]
	if userFilter.color[name] ~= nil then
		return "Filter already exists: "..name
	end
	userFilter.color[name] = false
	userFilter.modified = true
	options:AddFilterWidget(map, name)
end

local function FilterAddButton_OnClick(widget, event, ...)
	HideUIPanel(ColorPickerFrame)
	local name = filterControls.nameEdit:GetText()
	if name and name ~= "" then
		local errorMsg = options:AddFilter(controls.filterZoneDropdown:GetValue(), name)
		if errorMsg then
			filterControls.nameEdit:SetLabel(errorMsg)
		else
			filterControls.nameEdit:SetText()
			C_Timer.After(0.3, RefreshFilterNpcNames)
		end
	end
end

function options:AddFilterButton()
	local nameEdit = GUI:Create("EditBox")
	nameEdit:SetRelativeWidth(0.5)
	nameEdit:SetLabel("Enter NPC ID or Name")
	filterControls.nameEdit = nameEdit
	controls.filterGroup:AddChild(nameEdit)

	local addButton = GUI:Create("Button")
	addButton:SetRelativeWidth(0.22)
	addButton:SetText(ADD_FILTER)
	addButton:SetCallback("OnClick", FilterAddButton_OnClick)
	filterControls.addButton = addButton
	controls.filterGroup:AddChild(addButton)
end

function options:RemoveFilterWidget(map, name)
	for k, v in pairs(filterControls) do
		if k:find(name.."\\", 1, true) == 1 then
			filterControls[k] = nil
			v:Release()
			tDeleteItem(controls.filterGroup.children, v)
		end
	end
end

if not mainPanel:GetParent() then
	InterfaceOptions_AddCategory(mainPanel)
end
InterfaceOptions_AddCategory(filterPanel)
InterfaceAddOnsList_Update()
--InterfaceOptionsFrame_OpenToCategory(filterPanel)

mainPanel:refresh()
filterPanel:Hide()
--filterPanel:refresh()
C_Timer.After(0.5, filterPanel.refresh)
